import unittest
import pandas as pd
from ddi_fw.vectorization import SimilarityMatrixGenerator, VectorGenerator
 
 ## TODO: Complete the test case below

from sklearn.preprocessing import MultiLabelBinarizer

 

class TestSimilarityMatrixGenerator(unittest.TestCase):
    def setUp(self):
        # Create a sample dataframe
        data = {
            'column1': [[1, 0, 1], [1, 1, 0], [0, 0, 1]],
            'column2': [[0, 1, 0], [1, 0, 1], [1, 1, 1]]
        }
        self.dataframe = pd.DataFrame(data)

    def test_compute_jaccard_similarity(self):
        
        # Create an instance of SimilarityMatrixGenerator
        generator = SimilarityMatrixGenerator()

        # Compute the Jaccard similarity matrix
        columns = ['column1', 'column2']
        
        vector_generator = VectorGenerator(self.dataframe)
        generated_vectors = vector_generator.generate_feature_vectors(
        columns)
        

        # Print the similarity matrices
        for column, matrix in zip(columns, generated_vectors.values()):
            print(f"Jaccard similarity matrix for {column}:")
            print(matrix)
            sm = generator.create_jaccard_similarity_matrices(matrix)
            print(sm)

        # # Assertions to verify the results
        # self.assertIn('column1', similarity_matrices)
        # self.assertIn('column2', similarity_matrices)
        # self.assertIsInstance(similarity_matrices['column1'], pd.DataFrame)
        # self.assertIsInstance(similarity_matrices['column2'], pd.DataFrame)

if __name__ == '__main__':
    unittest.main()


import unittest
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
import numpy as np
from ddi_fw.ml import TFModelWrapper

import os

from ddi_fw.ml.evaluation_helper import Metrics

os.environ['TF_ENABLE_ONEDNN_OPTS'] = '0'

import keras

class TestTFModelWrapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Load the Iris dataset
        iris = load_iris()
        cls.X = iris.data
        cls.y = iris.target.reshape(-1, 1)

        # One-hot encode the target variable
        encoder = OneHotEncoder(sparse_output=False)
        cls.y = encoder.fit_transform(cls.y)

        # Split the dataset into training and testing sets
        cls.X_train, cls.X_temp, cls.y_train, cls.y_temp = train_test_split(cls.X, cls.y, test_size=0.4, random_state=42)
        cls.X_valid, cls.X_test, cls.y_valid, cls.y_test = train_test_split(cls.X_temp, cls.y_temp, test_size=0.5, random_state=42)

        # Define a simple neural network model
        # cls.model = cls.create_model((cls.X_train.shape[1],))

    @staticmethod
    def create_model(input_shape):
        model = Sequential([
            Dense(128, activation='relu', input_shape=(input_shape[1],)),
            Dense(64, activation='relu'),
            Dense(3, activation='softmax')
        ])
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
        return model

    def test_fit_and_evaluate(self):
        # Create an instance of TFModelWrapper
        tf_model_wrapper = TFModelWrapper(
            date='2021-09-01',
            model_func=self.create_model,
            descriptor='iris_test',
            use_mlflow=False
        )
        #train_idx_arr, val_idx_arr, train_data, train_label, test_data, test_label
        tf_model_wrapper.set_data(
            None,
            None,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test
        )
        # # Fit the model
        # best_model, best_model_key = tf_model_wrapper.fit()


        # # Predict using the best model
        # predictions = tf_model_wrapper.predict()

        # Evaluate the model
        logs, metrics, pred = tf_model_wrapper.fit_and_evaluate(print_detail=True)

        # Assertions to verify the results
        # self.assertIsNotNone(best_model)
        # self.assertIsInstance(best_model_key, str)
        # self.assertIsNotNone(predictions)
        self.assertIsInstance(logs, dict)
        self.assertIsInstance(metrics, Metrics)
        self.assertIsNotNone(pred)
        print(metrics)

if __name__ == '__main__':
    unittest.main()


from abc import ABC, abstractmethod
from typing import List, Any, Optional, Dict
import numpy as np
from ddi_fw.ml.model_wrapper import ModelWrapper, Result
from ddi_fw.ml.evaluation_helper import evaluate, Metrics
from ddi_fw.utils import utils
from ddi_fw import utils

import tensorflow as tf


def convert_to_categorical(arr, num_classes):
    """
    This function takes an array of labels and converts them to one-hot encoding 
    if they are not binary-encoded. If the array is already in a 
    compatible format, it returns the original array.

    Parameters:
    - arr: numpy array with label data (could be binary-encoded or label-encoded)
    - num_classes: number of classes to be used in one-hot encoding

    Returns:
    - The one-hot encoded array if the original array was binary or label encoded
    - The original array if it doesn't require any conversion
    """

    try:
        # First, check if the array is binary-encoded
        if not utils.is_binary_encoded(arr):
            # If the arr labels are binary-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(np.argmax(arr, axis=1), num_classes=num_classes)
        else:
            print("No conversion needed, returning original array.")
            return arr
    except Exception as e:
        # If binary encoding check raises an error, print it and continue to label encoding check
        print(f"Error while checking binary encoding: {e}")

    try:
        # Check if the array is label-encoded
        if utils.is_label_encoded(arr):
            # If the arr labels are label-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(arr, num_classes=num_classes)
    except Exception as e:
        # If label encoding check raises an error, print it
        print(f"Error while checking label encoding: {e}")
        # If the arr labels don't match any of the known encodings, raise an error
        raise ValueError("Unknown label encoding format.")

    # If no conversion was needed, return the original array

    return arr


class EnsembleStrategy(ABC):
    """Base class for ensemble strategies"""

    @abstractmethod
    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        """Combine individual predictions into ensemble prediction"""
        pass

    @abstractmethod
    def get_strategy_name(self) -> str:
        """Return strategy name"""
        pass


class VotingStrategy(EnsembleStrategy):
    """Majority voting strategy for classification"""

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        import numpy as np
        from collections import Counter
        predictions_array = np.array(predictions)
        pred_labels = np.array([p.argmax(axis=1) for p in predictions_array])
        return np.apply_along_axis(
            lambda x: Counter(x).most_common(1)[0][0],
            axis=0,
            arr=pred_labels
        )

    def get_strategy_name(self) -> str:
        return "voting"


class AveragingStrategy(EnsembleStrategy):
    """Averaging strategy for regression or probability outputs"""

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        return np.mean(np.array(predictions), axis=0)

    def get_strategy_name(self) -> str:
        return "averaging"


class StackingStrategy(EnsembleStrategy):
    """Stacking strategy using a meta-learner"""

    def __init__(self, meta_learner: ModelWrapper, base_wrappers: List[ModelWrapper] = []):
        self.meta_learner = meta_learner
        self.base_wrappers = base_wrappers
        self.descriptor = "stacking_ensemble"

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        # Stack predictions as features for meta-learner
        self.fit()
        x =  np.stack(predictions, axis=1)
        print("shape of predictions for stacking:", x.shape)
        self.meta_learner.test_data = x
        # self.meta_learner.fit(np.column_stack(predictions), None)
        # stacked_features = np.column_stack(predictions)
        return self.meta_learner.predict()

    def get_strategy_name(self) -> str:
        return "stacking"

    def fit(self):
         
        # preds = [b.predict() for b in self.base_wrappers]
        # self.meta_learner.test_data = np.stack(preds, axis=1) #test data should transformed to same shape of training data

        # self.meta_learner.test_data = np.hstack(
        #     [b.predict() for b in self.base_wrappers])

        print("Training meta-learner...")
        best_meta_model, best_meta_key, _ = self.meta_learner.fit()
        self.meta_learner.best_model = best_meta_model
        
        # self.best_model = best_meta_model

        return best_meta_model, best_meta_key

    def predict(self):
        # Step 3: Feed into meta-model
        pred = self.meta_learner.predict()
        return pred

    # def fit_and_evaluate(self, print_detail=False):
    #     best_model, best_key = self.fit()
    #     pred = self.predict()
    #     print("Stacking Strategy Predictions:")
    #     print(pred)
    #     pred_as_cat = convert_to_categorical(pred, self.num_classes)
    #     logs, metrics = evaluate(
    #         self.meta_learner.test_label, pred_as_cat, info=self.descriptor, print_detail=print_detail)
    #     metrics.format_float()
    #     metrics.set_time(self.meta_learner.elapsed_time)
    #     return logs, metrics, pred


class GenericEnsembleWrapper(ModelWrapper):
    """Generic ensemble wrapper supporting multiple strategies"""

    STRATEGY_MAP = {
        "voting": VotingStrategy,
        "averaging": AveragingStrategy,
        "stacking": StackingStrategy
    }

    def __init__(
        self,
        base_wrappers: List[ModelWrapper],
        ensemble_strategies: str,          # now comma separated input
        name: str = "ensemble",
        meta_learner: Optional[ModelWrapper] = None,
        tracking_service: Optional[Any] = None,
        date: Optional[str] = None
    ):
        self.base_wrappers = base_wrappers
        self.name = name
        self.tracking_service = tracking_service
        self.date = date or utils.utc_time_as_string_simple_format()
        self.ensemble_strategies = ensemble_strategies
        self.meta_learner = meta_learner

        # prediction cache so we don't recompute
        self.wrapper_predictions: Dict[str, np.ndarray] = {}

        self.individual_metrics = {}
        self.ensemble_metrics: Dict[str, Metrics] = {}
        self.result = Result()

    # ---------------------------------------------------------
    # FIT
    # ---------------------------------------------------------
    def fit(self):
        train_label, test_label = None, None

        # if any(getattr(wrapper, 'best_model', None) is not None for wrapper in self.base_wrappers):
        #     # At least one wrapper has best_model set
        #     pass
        
        train_label, test_label = None, None
        train_idx_arr, val_idx_arr = None, None
        val_labels= []
        inputs = []

        for wrapper in self.base_wrappers:
            train_label = wrapper.train_label
            test_label = wrapper.test_label

            train_idx_arr = wrapper.train_idx_arr
            val_idx_arr = wrapper.val_idx_arr
            

            if not val_labels:
              for i, (train_idx, val_idx) in enumerate(zip(train_idx_arr, val_idx_arr)):
                fold_val_labels = train_label[val_idx]
                val_labels.append(fold_val_labels)
            # Convert the list of arrays to a single numpy array
            all_fold_val_labels = np.concatenate(val_labels)
            
            print(f"Training base model: {wrapper.descriptor}")
            best_model, _, val_preds_dict = wrapper.fit()
            num_classes = wrapper.num_classes
            wrapper.best_model = best_model
            # pred_train = wrapper.predict()
            # if val_preds_dict is not None:
            print("val_preds_dict")
            print(val_preds_dict)
            stacked_values = np.concatenate(list(val_preds_dict.values()))
    
            # fold 1 training 800, validation 200 => training_val, validation_val
            # fold 2 training 800, validation 200
            # fold 3 training 800, validation 200
            # fold 4 training 800, validation 200
            # fold 5 training 800, validation 200
            
            # TRANSFORM
            # FOLD1.MODEL.PREDICT(TRAİNİNG1) => 800 
            # FOLD2.MODEL.PREDICT(TRAİNİNG2) => 800
            # FOLD3.MODEL.PREDICT(TRAİNİNG3) => 800
            # FOLD4.MODEL.PREDICT(TRAİNİNG4) => 800
            # FOLD5.MODEL.PREDICT(TRAİNİNG5) => 800
                
        inputs.append(stacked_values)

        self.num_classes = num_classes
        self.test_label = test_label
        print("val labels")
        print(len(val_labels))
        X_meta_train = np.stack(inputs, axis=1)
        # y_meta_train = train_label
        y_meta_train = all_fold_val_labels
        print("X_meta_train shape:", X_meta_train.shape)
        print(y_meta_train)
        print("y_meta_train shape:", y_meta_train.shape)

        # Step 3: Train meta-model
        self.meta_learner.train_data = X_meta_train
        self.meta_learner.train_label = y_meta_train
        self.meta_learner.train_idx_arr = train_idx_arr
        self.meta_learner.val_idx_arr = val_idx_arr
        self.meta_learner.test_label = test_label
        self.num_classes = num_classes

        # Parse comma-separated strategies
        self.selected_strategies: List[EnsembleStrategy] = []
        for s in self.ensemble_strategies.split(","):
            s = s.strip().lower()
            if s not in self.STRATEGY_MAP:
                raise ValueError(f"Unknown strategy '{s}'")
            if s == "stacking":
                if self.meta_learner is None:
                    raise ValueError(
                        "Meta-learner must be provided for stacking strategy")
                self.selected_strategies.append(StackingStrategy(
                    meta_learner=self.meta_learner, base_wrappers=self.base_wrappers))
            else:
                self.selected_strategies.append(self.STRATEGY_MAP[s]())

    # ---------------------------------------------------------
    # PREDICT — with prediction caching
    # ---------------------------------------------------------

    def _compute_and_cache_predictions(self):
        """Compute wrapper predictions only once."""
        if self.wrapper_predictions:
            return  # already computed

        for wrapper in self.base_wrappers:
            descriptor = wrapper.descriptor

            y_pred = wrapper.predict()
            print(f"Predictions from wrapper {descriptor}:")
            print(y_pred)
            print("Shape:")
            print(y_pred.shape)

            # Store raw predictions (strategies want raw data)
            self.wrapper_predictions[descriptor] = y_pred

            # Evaluation for individual wrapper
            y_pred_cat = convert_to_categorical(y_pred, self.num_classes)
            _, wrapper_metrics = evaluate(self.test_label, y_pred_cat)
            wrapper_metrics.set_time(wrapper.elapsed_time)
            wrapper_metrics.format_float()

            self.result.add_metric(descriptor, wrapper_metrics)
            self.individual_metrics[descriptor] = wrapper_metrics
        print(self.wrapper_predictions)
    # ---------------------------------------------------------
    # MAIN PREDICTION PIPELINE
    # ---------------------------------------------------------
    def predict(self):
        # Compute predictions only once
        self._compute_and_cache_predictions()

        pred_list = list(self.wrapper_predictions.values())

        ensemble_outputs = {}

        for strategy in self.selected_strategies:
            strategy_name = strategy.get_strategy_name()

            # Combine predictions
            ensemble_pred = strategy.combine_predictions(pred_list)
            ensemble_pred = convert_to_categorical(
                ensemble_pred, self.num_classes)

            # Evaluate ensemble
            _, metrics = evaluate(self.test_label, ensemble_pred)
            metrics.format_float()

            token = f"ensemble_{strategy_name}"
            self.ensemble_metrics[token] = metrics
            self.result.add_metric(token, metrics)

            ensemble_outputs[token] = ensemble_pred

        return ensemble_outputs   # multiple ensemble predictions

    # ---------------------------------------------------------
    # FIT AND EVALUATE
    # ---------------------------------------------------------
    def fit_and_evaluate(self):
        self.fit()
        ensemble_preds = self.predict()
        return None, self.result, ensemble_preds
 


import json
from typing import Optional
from ddi_fw.pipeline.pipeline import Pipeline
from ddi_fw.pipeline.ner_pipeline import NerParameterSearch
import importlib


def load_config(file_path):
    with open(file_path, 'r') as file:
        config = json.load(file)
    return config


def get_import(full_path_of_import):
    """Dynamically imports an object from a module given its full path.

    Args:
        full_path_of_import (str): The full path of the import (e.g., 'module.submodule.ClassName').

    Returns:
        object: The imported object.

    Raises:
        ImportError: If the module cannot be imported.
        AttributeError: If the attribute does not exist in the module.
    """
    if not full_path_of_import:
        raise ValueError("The import path cannot be empty.")

    parts = full_path_of_import.split('.')
    import_name = parts[-1]
    module_name = ".".join(parts[:-1]) if len(parts) > 1 else ""

    try:
        module = importlib.import_module(module_name)
        return getattr(module, import_name)
    except ModuleNotFoundError as e:
        raise ImportError(f"Module '{module_name}' could not be found.") from e
    except AttributeError as e:
        raise AttributeError(
            f"'{module_name}' has no attribute '{import_name}'") from e


class MultiPipeline():
    # def __init__(self, experiments_config_file, experiments_config):
    #     if experiments_config_file is None and experiments_config is None:
    #         raise ValueError("Either experiments_config_file or experiments_config must be provided.")
    #     if experiments_config_file is not None and experiments_config is not None:  
    #         raise ValueError("Only one of experiments_config_file or experiments_config should be provided.")
    #     if experiments_config_file is not None:
    #         self.experiments_config = load_config(experiments_config_file)
    #     else:
    #         self.experiments_config = experiments_config
    #     self.items = []
    #     self.pipeline_resuts = dict()
    
    def __init__(self, experiments_config_file: Optional[str] = None, experiments_config: Optional[dict] = None):
        """
        Initialize the MultiPipeline.

        Args:
            experiments_config_file (str, optional): Path to the experiments configuration file.
            experiments_config (dict, optional): Dictionary containing the experiments configuration.

        Raises:
            ValueError: If neither or both of the parameters are provided.
        """
        self.experiments_config = self._validate_and_load_config(experiments_config_file, experiments_config)
        self.items = []
        # self.pipeline_results = {}
        self.pipeline_resuts = dict()

    def _validate_and_load_config(self, experiments_config_file: Optional[str], experiments_config: Optional[dict]) -> dict:
        """
        Validate and load the experiments configuration.

        Args:
            experiments_config_file (str, optional): Path to the experiments configuration file.
            experiments_config (dict, optional): Dictionary containing the experiments configuration.

        Returns:
            dict: The loaded experiments configuration.

        Raises:
            ValueError: If neither or both of the parameters are provided.
        """
        if experiments_config_file is None and experiments_config is None:
            raise ValueError("Either 'experiments_config_file' or 'experiments_config' must be provided.")
        if experiments_config_file is not None and experiments_config is not None:
            raise ValueError("Only one of 'experiments_config_file' or 'experiments_config' should be provided.")

        if experiments_config_file is not None:
            try:
               config = load_config(experiments_config_file)
            except FileNotFoundError:
                raise FileNotFoundError(f"Configuration file '{experiments_config_file}' not found.")
        else:
            config = experiments_config
        if config is None:
            raise ValueError("Configuration cannot be None.")
        if not isinstance(config, dict):
            raise ValueError("Configuration must be a dictionary.")
        # if "experiments" not in config: 
        #     raise ValueError("Configuration must contain 'experiments' key.")
        return config

    def __create_pipeline(self, config):
        type = config.get("type")
        library = config.get("library")

        use_mlflow = config.get("use_mlflow")
        experiment_name = config.get("experiment_name")
        experiment_description = config.get("experiment_description")
        experiment_tags = config.get("experiment_tags")
        tracking_uri = config.get("tracking_uri")
        artifact_location = config.get("artifact_location")
        #new
        default_model = config.get("default_model")
        multi_modal = config.get("multi_modal")
        columns = config.get("columns")
        ner_data_file = config.get("ner_data_file")
        ner_threshold = config.get("ner_threshold")
        ner_min_threshold_dict = config.get("ner_min_threshold_dict")
        ner_max_threshold_dict = config.get("ner_max_threshold_dict")
        column_embedding_configs = config.get("column_embedding_configs")
        vector_db_persist_directory = config.get("vector_db_persist_directory")
        vector_db_collection_name = config.get("vector_db_collection_name")
        embedding_pooling_strategy = get_import(
            config.get("embedding_pooling_strategy_type")) if config.get("embedding_pooling_strategy_type") else None
        # Dynamically import the model and dataset classes
        # model_type = get_import(config.get("model_type"))
        dataset_type = get_import(config.get("dataset_type"))
        dataset_splitter_type = get_import(config.get("dataset_splitter_type"))

        combination_type = None
        kwargs_combination_params=None
        if config.get("combination_strategy"):
            combination_type = get_import(config.get("combination_strategy").get("type"))
            kwargs_combination_params = config.get("combination_strategy").get("params")
        combinations = []
        if combination_type is not None:
            combinations = combination_type(**kwargs_combination_params).generate()
      

        pipeline = None
        if type == "general":
            pipeline = Pipeline(
                library=library,
                use_mlflow=use_mlflow,
                experiment_name=experiment_name,
                experiment_description=experiment_description,
                experiment_tags=experiment_tags,
                artifact_location=artifact_location,
                tracking_uri=tracking_uri,
                dataset_type=dataset_type,
                dataset_splitter_type=dataset_splitter_type,
                columns=columns,
                column_embedding_configs=column_embedding_configs,
                vector_db_persist_directory=vector_db_persist_directory,
                vector_db_collection_name=vector_db_collection_name,
                embedding_pooling_strategy_type=embedding_pooling_strategy,
                ner_data_file=ner_data_file,
                ner_threshold=ner_threshold,
                combinations=combinations,
                default_model=default_model,
                multi_modal= multi_modal)
        elif type== "ner_search":
            pipeline = NerParameterSearch(
                library=library,
                experiment_name=experiment_name,
                experiment_description=experiment_description,
                experiment_tags=experiment_tags,
                tracking_uri=tracking_uri,
                dataset_type=dataset_type,
                dataset_splitter_type=dataset_splitter_type,
                umls_code_types = None,
                text_types = None,
                min_threshold_dict=ner_min_threshold_dict,
                max_threshold_dict=ner_max_threshold_dict,
                columns=columns,
                ner_data_file=ner_data_file,
                default_model=default_model,
                multi_modal= multi_modal
            )


        return {
            "name": experiment_name,
            "library": library,
            "pipeline": pipeline}

    def build(self):
        for config in self.experiments_config['experiments']:
            item = self.__create_pipeline(config)
            self.items.append(item)
        return self

    def run(self):
        for item in self.items:
            print(f"{item['name']} is running")
            pipeline = item['pipeline']
            pipeline.build()
            result = pipeline.run()
            self.pipeline_resuts[item['name']] = result
        return self

    def results(self):
        return self.pipeline_resuts


import numpy as np
from ddi_fw import utils
from ddi_fw.ml.model_wrapper import ModelWrapper
from ddi_fw.ml.evaluation_helper import Metrics, evaluate
import tensorflow as tf

def convert_to_categorical(arr, num_classes):
    """
    This function takes an array of labels and converts them to one-hot encoding 
    if they are not binary-encoded. If the array is already in a 
    compatible format, it returns the original array.

    Parameters:
    - arr: numpy array with label data (could be binary-encoded or label-encoded)
    - num_classes: number of classes to be used in one-hot encoding

    Returns:
    - The one-hot encoded array if the original array was binary or label encoded
    - The original array if it doesn't require any conversion
    """

    try:
        # First, check if the array is binary-encoded
        if not utils.is_binary_encoded(arr):
            # If the arr labels are binary-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(np.argmax(arr, axis=1), num_classes=num_classes)
        else:
            print("No conversion needed, returning original array.")
            return arr
    except Exception as e:
        # If binary encoding check raises an error, print it and continue to label encoding check
        print(f"Error while checking binary encoding: {e}")

    try:
        # Check if the array is label-encoded
        if utils.is_label_encoded(arr):
            # If the arr labels are label-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(arr, num_classes=num_classes)
    except Exception as e:
        # If label encoding check raises an error, print it
        print(f"Error while checking label encoding: {e}")
        # If the arr labels don't match any of the known encodings, raise an error
        raise ValueError("Unknown label encoding format.")

    # If no conversion was needed, return the original array

    return arr


class StackingWrapper(ModelWrapper):
    def __init__(self, date, descriptor, base_wrappers, meta_model_wrapper, tracking_service=None):
        """
        Args:
            base_wrappers (list): list of instantiated base model wrappers (TF or XGBoost)
            meta_model_wrapper (ModelWrapper): instantiated meta model wrapper
            tracking_service: optional tracking service
        """
        super().__init__(date, descriptor, None)
        self.base_wrappers = base_wrappers
        self.meta_model_wrapper = meta_model_wrapper
        self.tracking_service = tracking_service

    def fit(self):
        # Step 1: Train base models and collect softmax outputs
        inputs = []
        train_label, test_label = None, None
        train_idx_arr, val_idx_arr = None, None
        num_classes = None
        for wrapper in self.base_wrappers:
            train_label = wrapper.train_label
            test_label = wrapper.test_label
            
            # wrapper.train_data = self.train_data
            # wrapper.train_label = self.train_label
            train_idx_arr = wrapper.train_idx_arr
            val_idx_arr = wrapper.val_idx_arr
            # wrapper.train_idx_arr = self.train_idx_arr
            # wrapper.test_data = self.test_data
            # wrapper.test_label = self.test_label

            print(f"Training base model: {wrapper.descriptor}")
            best_model, _, val_preds_dict = wrapper.fit()
            num_classes = wrapper.num_classes
            wrapper.best_model = best_model
            # pred_train = wrapper.predict()
            # if val_preds_dict is not None:
            print("val_preds_dict")
            print(val_preds_dict)
            stacked_values = np.concatenate(list(val_preds_dict.values()))
            # Use validation predictions for training meta-model
            inputs.append(stacked_values)

            # pred_val = wrapper.predict()
            # base_test_outputs.append(pred_val)

        # Step 2: Prepare meta-model input
        # shape: (n_samples, num_classes * num_base_models)
        # X_meta_train = np.hstack(inputs)
        X_meta_train = np.stack(inputs, axis=1)
        y_meta_train = train_label

        # Step 3: Train meta-model
        self.meta_model_wrapper.train_data = X_meta_train
        self.meta_model_wrapper.train_label = y_meta_train
        self.meta_model_wrapper.train_idx_arr = train_idx_arr
        self.meta_model_wrapper.val_idx_arr = val_idx_arr
        self.num_classes = num_classes
        # self.meta_model_wrapper.val_idx_arr = None
        preds = [b.predict() for b in self.base_wrappers]
        self.meta_model_wrapper.test_data = np.stack(preds, axis=1) 
        
        # self.meta_model_wrapper.test_data = np.hstack(
        #     [b.predict() for b in self.base_wrappers])
        
        self.meta_model_wrapper.test_label = test_label

        print("Training meta-learner...")
        best_meta_model, best_meta_key, _ = self.meta_model_wrapper.fit()
        self.meta_model_wrapper.best_model = best_meta_model
        # self.best_model = best_meta_model

        return best_meta_model, best_meta_key

    def predict(self):
        # Step 3: Feed into meta-model
        pred = self.meta_model_wrapper.predict()
        return pred

    def fit_and_evaluate(self, print_detail=False):
        best_model, best_key = self.fit()
        pred = self.predict()
        pred_as_cat = convert_to_categorical(pred, self.num_classes)
        logs, metrics = evaluate(
            self.meta_model_wrapper.test_label, pred_as_cat, info=self.descriptor, print_detail=print_detail)
        metrics.format_float()
        return logs, metrics, pred



from typing import Any, Dict, List, Tuple
import time

from ddi_fw.ml.evaluation_helper import Metrics

class Result:
    def __init__(self) -> None:
        self.log_dict = {}
        self.metric_dict = {}

    def add_log(self, key, logs):
        self.log_dict[key] = logs

    def add_metric(self, key, metrics):
        self.metric_dict[key] = metrics


class ModelWrapper:
    def __init__(self, date, descriptor, model_func ,**kwargs):
        self.date = date
        self.descriptor = descriptor
        self.model_func = model_func
        self.kwargs = kwargs
        self.elapsed_time: float | None = None
       

    def set_data(self, train_idx_arr, val_idx_arr, train_data, train_label, test_data, test_label):
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr
        self.train_data = train_data
        self.train_label = train_label
        self.test_data = test_data
        self.test_label = test_label

    def _predict(self, model, X)-> Any:
        pass
    
    def predict(self)-> Any:
        pass

    def fit_model(self, X_train, y_train, X_valid, y_valid)-> Any:
        pass
    
    def fit(self)-> Any:
        pass
    
        # 🧠 Automatically wraps fit_model() when subclassing
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if hasattr(cls, "fit"):
            original_method = cls.fit

            def timed_fit(self, *args, **kwargs):
                start = time.perf_counter()
                result = original_method(self, *args, **kwargs)
                self.elapsed_time = time.perf_counter() - start
                return result

            cls.fit = timed_fit


import importlib


def get_import(full_path_of_import):
    """Dynamically imports an object from a module given its full path.

    Args:
        full_path_of_import (str): The full path of the import (e.g., 'module.submodule.ClassName').

    Returns:
        object: The imported object.

    Raises:
        ImportError: If the module cannot be imported.
        AttributeError: If the attribute does not exist in the module.
    """
    if not full_path_of_import:
        raise ValueError("The import path cannot be empty.")

    parts = full_path_of_import.split('.')
    import_name = parts[-1]
    module_name = ".".join(parts[:-1]) if len(parts) > 1 else ""

    try:
        module = importlib.import_module(module_name)
        return getattr(module, import_name)
    except ModuleNotFoundError as e:
        raise ImportError(f"Module '{module_name}' could not be found.") from e
    except AttributeError as e:
        raise AttributeError(
            f"'{module_name}' has no attribute '{import_name}'") from e


import unittest
from unittest import TestCase
import pandas as pd
from ddi_fw.vectorization import IDF


class IDFTesting(TestCase):
    def test_calculate_idf_values(self):
        item1 = 'T001|T002|T001|T001'
        item2 = 'T002|T003'
        item3 = 'T004|T005'

        all_data = [item1, item2, item3]

        df = pd.DataFrame(all_data, columns=['tui_description'])

        df['tui_description'] = df['tui_description'].apply(
            lambda x: x.split('|') if x is not None else [])

        print(df.head())

        idf = IDF(dataframe=df, columns=["tui_description"])
        idf.calculate()
        idf_scores_dict = idf.idf_scores
        idf_scores = idf.to_dataframe()
        print(idf_scores.head())

        # # idf_scores = calculate_idf(df['tui_description'])
        # idf_scores_sorted_desc = sorted(
        #     idf_scores.items(), key=lambda x: x[1], reverse=True)
        # threshold = 1
        # keys_over_threshold = [k for k,v in idf_scores.items() if v > threshold]


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
import math
from ddi_fw.utils.utils import clear_directory, create_folder_if_not_exists
import py7zr
import os
from os.path import basename

# https://py7zr.readthedocs.io/en/latest/user_guide.html
# import multivolumefile
#https://github.com/miurahr/py7zr/issues/497
class Py7ZipHelper:
    def __init__(self):
        pass

    def create_archive_from_file(self, archive_name, file_path, output_path):
        with py7zr.SevenZipFile(output_path+'/'+archive_name, 'w') as archive:
            # for file in files_to_archive:
            archive.write(file_path)

    def create_archive_from_folder(self, archive_name, folder_path, output_path):
        with py7zr.SevenZipFile(output_path+'/'+archive_name, 'w') as archive:
            for root, dirs, files in os.walk(folder_path):
                for file in files:
                    archive.write(os.path.join(root, file),
                                  os.path.relpath(os.path.join(root, file),
                                                  os.path.join(folder_path, '..')))
                    # archive.write(root+"/"+file)

    def create_archive_multiparts(self, zip_name, file_path, output_path, chunk_size, delete_existing_files=True):
        parent_folder = os.path.dirname(file_path)
        if delete_existing_files:
            clear_directory(output_path)
        # parts_path = f"{parent_folder}/parts"
        create_folder_if_not_exists(output_path)
        # file_name, file_extension = os.path.splitext(file_path)
        # file_name = os.path.basename(file_path)
        # file_name, folder = get_file_name_and_folder(file_path)

        if os.path.isdir(file_path):
            self.create_archive_from_folder(zip_name, file_path, output_path)
        elif os.path.isfile(file_path):
            self.create_archive_from_file(zip_name, file_path, output_path)
        else:
            return
        with open(output_path+'/'+zip_name, 'rb') as f:
            chunk_number = 1
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                with open(f"{output_path}/{zip_name}.part{chunk_number:03}", 'wb') as chunk_file:
                    chunk_file.write(chunk)
                chunk_number += 1
        if os.path.exists(output_path+'/'+zip_name):
            os.remove(output_path+'/'+zip_name)

    def create_archive(self, zip_prefix, input_path, output_path, chunk_size):
        files_paths = [input_path+'/' + p for p in os.listdir(input_path)]
        count_of_chunks = math.ceil(len(files_paths) / chunk_size)
        zero_padding_length = len(str(int(count_of_chunks))) + 2

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        part = 1
        i = 0
        zip_dict = defaultdict(list)
        for filePath in files_paths:
            padded_part = f'{part}'.zfill(zero_padding_length)
            key = f'{zip_prefix}.{padded_part}'
            zip_dict[key].append(filePath)
            i += 1
            if i % chunk_size == 0:
                i = 0
                part += 1

        for key, value in zip_dict.items():
            with py7zr.SevenZipFile(f'{output_path}/{key}.7z', 'w') as archive:
                for file_path in value:
                    archive.write(file_path, basename(file_path))

    def extract(self, input_path, output_path):
        files_paths = [input_path+'/' + p for p in os.listdir(input_path)]
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        for file_path in files_paths:
            if file_path.endswith('7z'):
                with py7zr.SevenZipFile(file_path, 'r') as z1:
                    z1.extractall(path=output_path)
                    print(f'{file_path} has been extracted')

    def extract_archive(self, archive_name, extract_path):
        with py7zr.SevenZipFile(archive_name, 'r') as archive:
            archive.extractall(path=extract_path)

    def extract_multiparts(self, input_path, output_path, output_file, remove_file = True):
        parts = [input_path+'/' + p for p in os.listdir(input_path)]
        sorted_parts = sorted(parts, key = lambda x: int(x.split(".")[-1][4:]))
        create_folder_if_not_exists(output_path)
        with open(f"{output_path}/{output_file}", 'wb') as outfile:
            for part in sorted_parts:
                with open(part, 'rb') as infile:
                    outfile.write(infile.read())
        self.extract_archive(f"{output_path}/{output_file}", output_path)
        if remove_file:
            os.remove(f"{output_path}/{output_file}")




from unittest import TestCase
from ddi_fw.pipeline.multi_modal_combination_strategy import CustomCombinationStrategy


class TryTesting(TestCase):
    
    def test_pure_combinations(self):
        import itertools
        l = ['e1','e2','e3','e4','e5']
        all_combinations = []
        for i in range(2, len(l) + 1):
            all_combinations.extend(list(itertools.combinations(l, i)))

        print(all_combinations)

        for combination in all_combinations:
            combination_descriptor = '-'.join(combination)
            print(combination_descriptor)
    
    def test_combination_strategy(self):
        group_1 = ["a", "b", "c"]
        group_2 = [1, 2, 3]
        strategy = CustomCombinationStrategy(group_1=group_1, group_2=group_2)
        combinations = strategy.generate()
        print(combinations)
